"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Processor = void 0;
const State_1 = require("../nodes/State");
const Operations_1 = require("./Operations");
const Util_1 = require("../Util");
const consts_1 = require("../nodes/consts");
const consts_2 = require("../consts");
class Processor {
    constructor(Nodes, variables, functions) {
        this.output = [];
        this.pathStringSteps = {
            item: null,
            nodeName: null,
            init: (item) => {
                this.pathStringSteps.item = item;
            },
            resetNodeName: () => {
                this.pathStringSteps.nodeName = null;
            },
            setNodeName: (step, subPath) => {
                if (step.includes('::')) {
                    const axisNode = subPath.split("::");
                    step = axisNode[0];
                    this.pathStringSteps.nodeName = axisNode[1];
                }
                return step;
            },
            // process each axis in an array of steps eg ['/', 'a', '//', 'b']
            // 'a' and 'b' would fall under 'default' axis
            axis: {
                '/': () => {
                    const item = this.pathStringSteps.item;
                    item.value = item.value.reduce((r, v) => {
                        const children = Util_1.nodesOps.get.children(v);
                        r = [...r, ...children];
                        return r;
                    }, []);
                },
                '//': () => {
                    const item = this.pathStringSteps.item;
                    item.value = item.value.reduce((r, v) => {
                        const descendants = Util_1.nodesOps.get.descendants(v);
                        r = [...r, ...descendants];
                        return r;
                    }, []);
                },
                '.': () => {
                },
                '..': () => {
                    const item = this.pathStringSteps.item;
                    item.value = item.value.reduce((r, v) => {
                        const parent = Util_1.nodesOps.get.parent(v);
                        if (parent && !r.some((rr) => rr[consts_1.KEYS.id] === parent[consts_1.KEYS.id])) {
                            r.push(parent);
                        }
                        return r;
                    }, []);
                },
                '*': () => { },
                parent: () => {
                    const item = this.pathStringSteps.item;
                    item.value = item.value.reduce((r, v) => {
                        const parent = Util_1.nodesOps.get.parent(v, this.pathStringSteps.nodeName);
                        if (parent && !r.some(node => node[consts_1.KEYS.id] === parent[consts_1.KEYS.id])) {
                            r.push(parent);
                        }
                        return r;
                    }, []);
                    this.pathStringSteps.resetNodeName();
                },
                child: () => {
                    const item = this.pathStringSteps.item;
                    item.value = item.value.reduce((r, v) => {
                        const children = Util_1.nodesOps.get.children(v, this.pathStringSteps.nodeName);
                        children.forEach(a => {
                            if (!r.some(node => node[consts_1.KEYS.id] === a[consts_1.KEYS.id])) {
                                r.push(a);
                            }
                        });
                        return r;
                    }, []);
                    this.pathStringSteps.resetNodeName();
                },
                ancestor: () => {
                    const item = this.pathStringSteps.item;
                    item.value = item.value.reduce((r, v) => {
                        const ancestors = Util_1.nodesOps.get.ancestors(v, this.pathStringSteps.nodeName);
                        ancestors.forEach(a => {
                            if (!r.some(node => node[consts_1.KEYS.id] === a[consts_1.KEYS.id])) {
                                r.push(a);
                            }
                        });
                        return r;
                    }, []);
                    this.pathStringSteps.resetNodeName();
                },
                'ancestor-or-self': () => {
                    const item = this.pathStringSteps.item;
                    item.value = item.value.reduce((r, v) => {
                        const ancestors = Util_1.nodesOps.get.ancestors(v, this.pathStringSteps.nodeName);
                        ancestors.forEach(a => {
                            if (!r.some(node => node[consts_1.KEYS.id] === a[consts_1.KEYS.id])) {
                                r.push(a);
                            }
                        });
                    }, [item.value]);
                    this.pathStringSteps.resetNodeName();
                },
                descendant: () => {
                    const item = this.pathStringSteps.item;
                    item.value = item.value.reduce((r, v) => {
                        const descendants = Util_1.nodesOps.get.descendants(v, this.pathStringSteps.nodeName);
                        r = [...r, ...descendants];
                        return r;
                    }, []);
                    this.pathStringSteps.resetNodeName();
                },
                'descendant-or-self': () => {
                    const item = this.pathStringSteps.item;
                    item.value = item.value.reduce((r, v) => {
                        const descendants = Util_1.nodesOps.get.descendants(v, this.pathStringSteps.nodeName);
                        r = [...r, ...descendants];
                        return r;
                    }, [item.value]);
                    this.pathStringSteps.resetNodeName();
                },
                self: () => {
                    const item = this.pathStringSteps.item;
                    const from = item.value;
                    item.value = Util_1.nodesOps.get.byName(item.value, this.pathStringSteps.nodeName);
                    this.pathStringSteps.resetNodeName();
                },
                sibling: () => {
                    const item = this.pathStringSteps.item;
                    item.value = item.value.reduce((r, v) => {
                        const siblings = Util_1.nodesOps.get.siblings(v, this.pathStringSteps.nodeName);
                        r = [...r, ...siblings];
                        return r;
                    }, []);
                    this.pathStringSteps.resetNodeName();
                },
                default: (value) => {
                    const item = this.pathStringSteps.item;
                    const initialList = item.value;
                    item.value = Util_1.nodesOps.get.byName(initialList, value);
                }
            }
        };
        this.tests = {
            isPositiveFilteredResult: (item) => {
                switch (item.type) {
                    case consts_2.TYPES.nodes:
                        return !!item.value.length;
                    case consts_2.TYPES.boolean:
                        return !!item.value;
                    case consts_2.TYPES.invalid:
                        return false;
                    default: debugger; /* TODO: handle other types */
                }
            },
            isPositionCheck: (paths) => {
                return Array.isArray(paths) && ((paths.length === 1 && paths[0].type === consts_2.TYPES.number) ||
                    (paths.length === 2 && paths[1].type === consts_2.TYPES.function && ['last', 'first'].includes(paths[1].value)));
            }
        };
        this.postFixPathTypes = {
            rootPath: (path, { isFilterMode }) => {
                this.output.push({
                    type: consts_2.TYPES.nodes,
                    value: Util_1.nodesOps.get.root(this.nodesState.getNodes(path.callerId)),
                    startedFromRoot: !!isFilterMode
                });
                this.processPathString(path.value);
            },
            path: (path, { isFilterMode, current }) => {
                if (isFilterMode && current && path.value[0] !== '/') {
                    this.output.push({
                        type: consts_2.TYPES.nodes,
                        value: [current]
                    });
                }
                // expects pop to pull out nodes type
                this.processPathString(path.value);
            },
            string: (path) => {
                this.output.push(path);
            },
            number: (path) => {
                this.output.push(path);
            },
            boolean: (path) => {
                this.output.push(path);
            },
            variablePath: (path) => {
                const variableName = path.value.substring(0, path.value.indexOf('/'));
                this.output.push({
                    type: consts_2.TYPES.nodes,
                    value: this.variablesInstance.getVariableRootNode(variableName)
                });
                this.processPathString(path.value.replace(variableName, ''));
            },
            variable: (path) => {
                this.output.push({
                    type: consts_2.TYPES.nodes,
                    value: this.variablesInstance.getVariableRootNode(path.value)
                });
                this.processPathString('.');
            },
            position: (path, { index }) => {
                this.output.push({
                    type: path.type,
                    value: index
                });
            },
            function: (path, filterProps) => {
                const args = this.output.pop();
                this.functionsInstance.preRun(filterProps);
                this.output.push(this.functionsInstance.run(path.value, args.value));
            },
            arguments: (path, filterOptions) => {
                this.output.push({
                    type: consts_2.TYPES.arguments,
                    value: path.value.map(arg => {
                        this.processPostfixPath(Array.isArray(arg) ? arg : [arg], filterOptions);
                        return this.output.pop();
                    })
                });
            },
            operator: (path, filterOptions) => {
                const rightOperandItem = this.output.pop();
                const leftOperandItem = this.output.pop();
                const result = this.operationsInstance.run(path.value, leftOperandItem, rightOperandItem, filterOptions);
                this.output.push(result);
            }
        };
        this.operationsInstance = new Operations_1.Operations();
        this.mainNodes = Nodes;
        this.variablesInstance = variables;
        this.functionsInstance = functions;
        this.nodesState = State_1.NodesState.getInstance();
    }
    ;
    processPathString(pathString) {
        let steps = [];
        let subPath = '';
        for (let i = 0; i < pathString.length; i++) {
            subPath += pathString[i];
            if ((subPath.includes('::') || subPath === '..') && steps[steps.length - 1] === '/') {
                // skip processing '/' step (children)
                steps.pop();
            }
            if (subPath !== '/' && pathString[i + 1] === '/') {
                steps.push(subPath);
                subPath = '';
            }
            else if (subPath === '/' && pathString[i + 1] !== '/' ||
                subPath === '//') {
                steps.push(subPath);
                subPath = '';
            }
            else if (i === pathString.length - 1) {
                steps.push(subPath);
            }
        }
        this.pathStringSteps.init(this.output.pop());
        while (steps.length) {
            let step = steps.shift();
            step = this.pathStringSteps.setNodeName(step, subPath);
            if (step.length && typeof this.pathStringSteps.axis[step] === 'function') {
                this.pathStringSteps.axis[step]();
            }
            else {
                this.pathStringSteps.axis['default'](step);
            }
        }
        this.output.push(this.pathStringSteps.item);
    }
    ;
    clonePath(originalPath) {
        return originalPath.map(path => {
            if (Array.isArray(path)) {
                return this.clonePath(path);
            }
            else {
                let value = path.value;
                return {
                    type: path.type,
                    value: Array.isArray(value) ? path.value.map(v => Array.isArray(v) ? [...v] : v) : value
                };
            }
        });
    }
    ;
    processPostfixPath(paths, filterProps) {
        let path;
        while (paths.length) {
            path = paths.shift();
            if (Array.isArray(path)) { // this triggers filter mode (xpath predicate)
                let isPositionCheck = false;
                const context = this.output.pop();
                const updatedNodes = context.value.filter((nodes, index) => {
                    // clone the path
                    const localPath = this.clonePath(path);
                    if (isPositionCheck || this.tests.isPositionCheck(localPath)) {
                        isPositionCheck = true;
                        localPath.push({
                            type: 'position',
                            value: ''
                        });
                        localPath.push({
                            type: 'operator',
                            value: '='
                        });
                    }
                    const localFilterProps = {
                        isFilterMode: true,
                        current: nodes,
                        index: index + 1,
                        firstIndex: 1,
                        lastIndex: context.value.length
                    };
                    this.processPostfixPath(localPath, localFilterProps);
                    return this.tests.isPositiveFilteredResult(this.output.pop());
                });
                this.output.push({
                    type: consts_2.TYPES.nodes,
                    value: updatedNodes
                });
            }
            else {
                this.postFixPathTypes[path.type](path, Object.assign({}, filterProps));
            }
        }
    }
    ;
    reset() {
        this.output = [];
        this.nodesState.reset();
    }
}
exports.Processor = Processor;
//# sourceMappingURL=Processor.js.map